// Copyright 2013-2014, SAS Institute Inc., Cary, NC, USA.  All Rights Reserved.

package com.sas.deployment
import groovy.transform.InheritConstructors

import com.sas.gate.CodeAccessor
import com.sas.gate.api.PipeCode
import com.sas.gate.api.template.*
import org.codehaus.groovy.runtime.StackTraceUtils

/**
   <overview>Deployment code for SAS Enterprise Case Management (12byte)</overview>
   <p>
      This code configures the deployment of 12byte using what corresponds to
      the following Ant targets:
      <ul class='nodot'>
         <li>configure</li>
         <li>unconfigure</li>
      </ul>
   </p>

   @since      14w14
   @version    14w14
   @author     Bruce Bryan
   @author     Steve Amerige
   @see        #SOURCE#
*/

@InheritConstructors
class CASEMGMTMVAC extends CodeAccessor
{
   // ===================================================== GLOBAL

   /**
      Self-identify this class name

      @param id <type>String</type>: The 12byte name
   */
   String id = 'CASEMGMTMVAC'

   /**
      Callback for hook: <hook>properties &gt; init</hook>

      @param t__properties_get   <type>ArrayList&lt;String&gt;</type>: get object properties on each dependson property

   <ul class='nodot'>
      <li>levconfig</li>
   </ul>
   */
   def properties_init =
   {
      pass.with {
         debug = ['debug']

         // Simple Strings
         _releaseNumber                         = api.'casemgmtmvac.release.number'
            assert _releaseNumber == "6.2" : "Release number does not match 6.2"
         _offeringLegalName                     = api.'casemgmtmvac.offering.legal.name'
            assert _offeringLegalName == 'SAS Enterprise Case Management' : "Offering legal name does not match SAS Enterprise Case Management"
         _legalName                             = api.'casemgmtmva.legal.name'
            assert _legalName == 'SAS Enterprise Case Management Server' : "Legal name does not match SAS Enterprise Case Management Server"
         _softwareComponentName                 = api.'casemgmtmvac.software.component.name'
            assert _softwareComponentName == 'Ent Case Mgmt Server 6.2' : "Software component name does not match Ent Case Mgmt Server 6.2"
         _description                           = "$_legalName $_releaseNumber"
            assert _description == 'SAS Enterprise Case Management Server 6.2' : "Description does not match SAS Enterprise Case Management Server 6.2"

         // Config Paths
         _configDir                             = api.'casemgmtmvac.config.dir'
            assert _configDir =~ "SASEnterpriseCaseManagement" : "Config dir does not contain SASEnterpriseCaseManagement"
            assert _configDir =~ "6.2" : "Config dir does not contain 6.2"
         _configSourceDir                       = canonical("$_configDir/Source")
            assert _configSourceDir =~ "Source" : "Config dir does not contain Source"
         _configLibrariesDir                    = canonical("$_configDir/Libraries")
            assert _configLibrariesDir =~ "Libraries" : "Config dir does not contain Libraries"
         _configCntlDir                         = canonical("$_configSourceDir/control")
            assert _configCntlDir =~ "control" : "Config dir does not contain control"
         _configTemplateDir                     = canonical("$_configSourceDir/form_template")
            assert _configTemplateDir =~ "form_template" : "Config dir does not contain form_template"
         _configSasmiscDir                      = canonical("$_configSourceDir/sasmisc")
            assert _configSasmiscDir =~ "sasmisc" : "Config dir does not contain sasmisc"
         _configStpDir                          = canonical("$_configSourceDir/sasstp")
            assert _configStpDir =~ "sasstp" : "Config dir does not contain sasstp"
         _configUcmacrosDir                     = canonical("$_configSourceDir/ucmacros")
            assert _configUcmacrosDir =~ "ucmacros" : "Config dir does not contain ucmacros"
         _configInstallDir                      = canonical("$_configSasmiscDir/install")
            assert _configInstallDir =~ "install" : "Config dir does not contain install"
         _configMetadataDir                     = canonical("${_configSasmiscDir}/install/metadata")
            assert _configMetadataDir =~ "metadata" : "Config dir does not contain metadata"
         _autoexecPath                          = canonical("$_configCntlDir/ecm_autoexec.sas")
            assert _autoexecPath =~ "ecm_autoexec.sas" : "Config dir does not contain ecm_autoexec.sas"

         // Install Paths (Dirs and Files)
         _installSASDir                         = api.'sasinstall.install.dir'
         _installConfigDir                      = api.'casemgmtmvac.install.config.dir'
         _installMetadataDir                    = canonical("$pass._installConfigDir/Deployment/Metadata")
         _installSASmiscDir                     = api.'casemgmtmvac.install.sasmisc.dir'
         _installSASstpDir                      = api.'casemgmtmvac.install.sasstp.dir'
         _installCmacrosDir                     = api.'casemgmtmvac.install.cmacros.dir'
         _installUcmacrosDir                    = api.'casemgmtmvac.install.ucmacros.dir'

         // Metadata Folders (Folders and Items)
         _applicationFolder                     = api.'casemgmtmvac.application.folder'
            assert _applicationFolder == '/System/Applications/SAS Enterprise Case Management' : "Application metadata folder does not match /System/Applications/SAS Enterprise Case Management"
         _applicationSubFolder                  = api.'casemgmtmvac.application.sub.folder'
            assert _applicationSubFolder == '/System/Applications/SAS Enterprise Case Management/Ent Case Mgmt Server 6.2' :
                                           "Application metadata sub folder does not match /System/Applications/SAS Enterprise Case Management/Ent Case Mgmt Server 6.2"
         _sasstpFolder                          = api.'casemgmtmvac.application.stp.folder'
            assert _sasstpFolder == '/System/Applications/SAS Enterprise Case Management/Ent Case Mgmt Server 6.2/Application SAS Code' :
                                   "Application metadata SAS folder does not match /System/Applications/SAS Enterprise Case Management/Ent Case Mgmt Server 6.2/Application SAS Code"
         _productLegalFolder                    = api.'casemgmtmvac.product.legal.folder'
            assert _productLegalFolder == '/Products/SAS Enterprise Case Management' : "Application metadata sub folder does not match /Products/SAS Enterprise Case Management"

         // Other
         _dbmsAuthdomainName                    = api.'casemgmtmvac.dbms.authdomain'
            assert _dbmsAuthdomainName == 'casemgmt_db_auth' : "casemgmtmvac.dbms.authdomain does not match casemgmt_db_auth"
         _dbmsUserPassword                      = api.'casemgmtmvac.dbms.passwd'
         _dbmsUserUserid                        = api.'casemgmtmvac.dbms.userid'
         _dbmsGroup                             = 'Ent Case Mgmt DB Group'
         _updateDatabaseName                    = api.'casemgmtmvac.dbms.updtedb.name'
         _dbmsDirLabel                          = api.'casemgmtmvac.dbms.dir.label'
         _dbmsEngine                            = api.'casemgmtmvac.dbms.engine'
         _databaseHost                          = api.'casemgmtmvac.dbms.host'
         _databasePort                          = api.'casemgmtmvac.dbms.port'
         _databaseName = api."${(_dbmsEngine == 'oracle') ? 'casemgmtmvac.dbms.oracle.sid': ((_dbmsEngine in ['odbc','sqlsvr']) ? 'casemgmtmvac.dbms.sqlserver.datasource.name' : 'casemgmtmvac.dbms.database.name')}"
         _maintenanceFrom = api.baseVersion(getProperty("casemgmtmvac.maintenance.from.version", ''))

         // Template APIs
         t__properties_get =
         [
            'casemgmtmvac.dependson.levconfig.fqid',
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param filters <type>Map&lt;String&gt;</type>: tokens for string substitution
      @param t__setobjectproperties    <type>ArrayList&lt;String&gt;</type>: Properties that will be stored in the configured objects text store
      @param t__createdirectories <type>Map&lt;String&gt;</type>: directories to be created in configure target<br>
            <indent><key>prefix</key>- <type>String</type>: base directory to create<br>
            <key>entries</key>- <type>ArrayList&lt;String&gt;</type>: directories and subdirectories to be created<br></indent>
      <pre>
   /install/config/Lev1/Applications/SASEnterpriseCaseManagement/6.2
|-- Libraries
|   `-- ecmreport
`-- Source
    |-- cmacros
    |-- control
    |-- efiles
    |-- form_template
    |-- sasmisc
    |   `-- install
    |       `-- Oracle
    |-- sasstp
    `-- ucmacros</pre>

      @param t__copyfiles         <type>ArrayList&lt;Map&gt;</type>: files to copy and conditions for the copy<br>
            <indent><key>src</key>- <type>String</type>: Source directory to copy from<br>
            <key>include</key>- <type>String</type>: Regex pattern of file(s) to copy<br>
            <key>dst</key>- <type>String</type>: Destination directory for files to be copied to<br>
            <key>strip</key>- <type>String</type> <kind>optional</kind>: Remove extension form file name i.e., remove .orig extension<br>
            <key>tokens</key>- <type>String</type> <kind>optional</kind>: filterset to use for token replacement<br></indent>
      @param t__loadmetadata_createfolders   -
      @param t__importpackages_createfolders   -
      @param t__loadmetadata_load   <type>String</type>: Fully qualified path to the token substituted appxml file
      @param t__sasjobs_run      -
   */
   def properties_post =
   {
      // SAS Job
      //def sasXXXJob                             = 'product_config'
      //def sasYYYJob                             = 'install'
      //def sasJobECMFailed                       = new PipeCode([api.'warnings_casemgmtmvac_loadInstallData_error.txt'])
      //   .tokenstringsub(tokens:['_!JOB_NAME':sasYYYJob],prefix:'${',suffix:'}').output[0]
      //def sasJobHPAMLFailed                     = new PipeCode([api.'warnings_casemgmtmvac_loadInstallData_error.txt'])
      //   .tokenstringsub(tokens:['_!JOB_NAME':sasXXXJob],prefix:'${',suffix:'}').output[0]

      // Properties
      api.createProperty('casemgmtmvac.config.dir', pass._configDir)
      api.createProperty('casemgmtmvac.autoexec.path', pass._autoexecPath)
      api.createProperty('casemgmtmvac.configobj.name', pass._softwareComponentName)

      // Other
      String classIdentifier                       = api.'casemgmtmvac.configobj.classid'
         assert classIdentifier == 'casemgmtmvac' : "Class identifier does not match casemgmtmvac"
      String majorNumber                           = api.'casemgmtmvac.major.number'
         assert majorNumber == '6' : "Release major number does not equal 6"
      String minorNumber                           = api.'casemgmtmvac.minor.number'
         assert minorNumber == '2' : "Release minor number does not equal 2"
      String serverContextName                     = api.'oma.appserver.servercontext.name'
      String authDomain                            = api.'oma.authenticationdomain.name'
      String metadataUser                          = api.'metadata.connection.userid'
      String cmDates                               = api.'casemgmtmvac.dbms.cmdates'
      String appserverContextDir                   = api.'config.lev.appservercontext.dir'
      String trustedUserPersonName                 = api.'oma.person.trustusr.person.name'
      String adminLoginPasswd                      = api.'metadata.connection.passwd'
      String metadataHost                          = api.'iomsrv.metadatasrv.host'
      String metadataPort                          = api.'iomsrv.metadatasrv.port'
      String databaseType                          = api.'casemgmtmvac.dbms.type'
      String repositoryName                        = api.'oma.repository.foundation.name'
      String midTierSoftwareComponentName          = api.'casemgmtmvac.casemgmtmid.software.component.name'

      pass.with
      {
         _filters =
         [
            'server.software.component.name'          : _softwareComponentName,
            'midtier.software.component.name'         : midTierSoftwareComponentName,
            'authdomain'                              : authDomain,
            'casemgmtmvac.config.dir.name'            : _configDir,
            'casemgmtmvac.configobj.name'             : _softwareComponentName,
            'casemgmtmvac.configobj.classid'          : classIdentifier,
            'casemgmtmvac.db.engine'                  : _dbmsEngine,
            'casemgmtmvac.offering.legal.name'        : _offeringLegalName,
            'cmdates'                                 : cmDates,
            'config.lev.appservercontext.dir'         : appserverContextDir,
            'configobj.classid'                       : classIdentifier,
            'configobj.name'                          : _softwareComponentName,
            'ecm_etl_output_path'                     : _configLibrariesDir,
            'major.number'                            : majorNumber,
            'minor.number'                            : minorNumber,
            'oma.appserver.servercontext.name'        : serverContextName,
            'oma.person.trustusr.person.name'         : trustedUserPersonName,
            'OMADUPW'                                 : adminLoginPasswd,
            'OMADUSER'                                : metadataUser,
            'OMAHOST'                                 : metadataHost,
            'OMAPORT'                                 : metadataPort,
            'rdbms.casemgmtmvac.auth'                 : _dbmsAuthdomainName,
            'rdbms.casemgmtmvac.connection.passwd'    : _dbmsUserPassword,
            'rdbms.casemgmtmvac.connection.userid'    : _dbmsUserUserid,
            'rdbms.casemgmtmvac.database.host'        : _databaseHost,
            'rdbms.casemgmtmvac.database.port'        : _databasePort,
            'rdbms.casemgmtmvac.database.name'        : _databaseName,
            'rdbms.casemgmtmvac.type'                 : _dbmsEngine,
            'REPOSITORY'                              : repositoryName,
            'SASROOT'                                 : _installSASDir,
            'server.entcmmva.storedprocess.targetdir' : _configStpDir.replaceAll('\\\\', '\\\\\\\\'),
            'server.casemgmtmvac.release.number'      : _releaseNumber,
            'server.casemgmtmvac.storedprocess.dir'   : _installSASstpDir,
            'casemgmtmvac.config.ucmacros.dir'        : _configUcmacrosDir,
            'casemgmtmvac.install.ucmacros.dir'       : _installUcmacrosDir,
            'casemgmtmvac.install.cmacros.dir'        : _installCmacrosDir,
            'casemgmtmvac.install.sasmisc.dir'        : _installSASmiscDir,
            'casemgmtmvac.config.control.dir'         : _configCntlDir
         ]

         t__createdirectories =
         [
            prefix: _configDir,
            entries:
            [
               'Libraries',
               'Libraries/ecmreport',
               'Source',
               'Source/cmacros',
               'Source/control',
               'Source/efiles',
               'Source/form_template',
               'Source/sasmisc',
               'Source/sasmisc/install',
               "Source/sasmisc/install/$_dbmsDirLabel" as String,
               'Source/sasmisc/install/metadata',
               'Source/sasstp',
               'Source/ucmacros'
            ]
         ]

         t__setobjectproperties =
         [
            properties:
            [
               'casemgmtmvac.config.dir',
               'casemgmtmvac.configobj.classid',
               'casemgmtmvac.configobj.order.number',
               'casemgmtmvac.configobj.version',
               'casemgmtmvac.dbms.userid',
               'casemgmtmvac.dbms.driver_class',
               'casemgmtmvac.dbms.url',
               'casemgmtmvac.dbms.database.host',
               'casemgmtmvac.dbms.database.name',
               'casemgmtmvac.dbms.database.port',
               'casemgmtmvac.dbms.db2.license.jar',
               'casemgmtmvac.dbms.jar',
               'casemgmtmvac.dbms.oracle.net.service.name',
               'casemgmtmvac.dbms.oracle.sid',
               'casemgmtmvac.dbms.schema',
               'casemgmtmvac.dbms.type',
               'casemgmtmvac.desc',
               'casemgmtmvac.install.config.dir',
               'casemgmtmvac.installdir.config.dir',
               'casemgmtmvac.legal.name',
               'casemgmtmvac.legal.short.name',
               'casemgmtmvac.offering.legal.name',
               'casemgmtmvac.release.number',
               'casemgmtmvac.short.legal.name',
               'casemgmtmvac.short.name',
               'casemgmtmvac.software.component.name',
               'casemgmtmvac.autoexec.path',
               'casemgmtmvac.configobj.name',
           ]
        ]

         t__copyfiles =
         [
            [  src:        "$_installSASmiscDir/control",
               include:    ~/.*\.orig/,
               dst:        _configCntlDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        "$_installSASmiscDir/sample/form_template",
               include:    ~/.*\.*/,
               dst:        _configTemplateDir,
            ],
            [  src:        "$_installSASmiscDir/install/ddl/$_updateDatabaseName",
               include:    ~/.*\.sql/,
               dst:        "$_configInstallDir/$_dbmsDirLabel",
               tokens:     _filters
            ],
            [  src:        "$_installSASmiscDir/install/config",
               include:    ~/.*\.(sas|sql)/,
               dst:        _configInstallDir,
               tokens:     _filters
            ],
            [  src:        _installSASstpDir,
               include:    ~/.*\.orig/,
               dst:        _configStpDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        _installUcmacrosDir,
               include:    ~/.*\.orig/,
               dst:        _configUcmacrosDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        _installMetadataDir,
               include:    ~/.*\.orig/,
               dst:        _configMetadataDir,
               strip:      '.orig',
               tokens:     _filters
            ],
            [  src:        _installMetadataDir,
               include:    ~/.*\.locprop/,
               exclude:    ~/.*\.orig/,
               dst:        _configMetadataDir,
            ],
         ]

         t__loadmetadata_createfolders =
         [
            _applicationFolder,
            _productLegalFolder
         ]

         t__loadmetadata_load = canonical("$_configMetadataDir/casemgmtmvac.appxml")

         if (pass.hasVariable('_updateMetadata') && _updateMetadata)
            updateconfigure__loadmetadata_init()

         t__importpackages_createfolders =
         [
            _sasstpFolder
         ]

         t__importpackages_run =
         [
            packages:
            [
               [
               name:          'ECMStoredProcesses',
               packagesDir:   canonical("$_installConfigDir/Deployment/Packages/"),
               folder:        _sasstpFolder,
               tokens:        _filters
               ]
            ]
         ]

         t__addlogintogroup =
         [
            authdomain: _dbmsAuthdomainName,
            login:      _dbmsUserUserid,
            passwd:     _dbmsUserPassword,
            group:      _dbmsGroup,
         ]

         t__setmetadataperms =
         [
            perm  : 'G',
            group : 'Ent Case Mgmt Users',
            set:
            [
               [
                  folders: _offeringLegalName,
                  parentfolder: 'Applications',
                  permNames:
                  [
                     'ReadMetadata','WriteMetadata','WriteMemberMetadata','CheckInMetadata',
                     'Read','Write','Create','Delete','Administer'
                  ]
               ],
               [
                  folders: _offeringLegalName,
                  parentfolder: 'Products',
                  permNames:
                  [
                     'ReadMetadata','WriteMetadata','WriteMemberMetadata','CheckInMetadata',
                     'Read','Write','Create','Delete','Administer'
                  ]
               ],
            ]
         ]

/*         t__sasjobs_run =
         [
            enabled: !t__migrate,
            folder:  "$configDir/install",
            groups:  [ '<GROUP_1>', '<GROUP_2>' ],
            jobs:
            [
               [
                  jobname: sasXXXJob,
                  autoexec: autoexecPath,
                  message: sasJobXXXFailed,
               ],
               [
                  jobname: sasYYYJob,
                  win:     [options:   '-CONFIG "' + installSASNLSConfigFile + '"'],
                  unx:     [sasexec:   installUnxU8sasexec],
                  message: sasJobYYYFailed,
               ]
            ]
         ] */
      }
   }

   // ===================================================== TARGET: configure

   /**
      Target: configure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def configure()
   {
      //api.stop()

      ConfigureTemplate t = new ConfigureTemplate(this)

      t['migration > init']            .add(configure__migration_init)
      t['migration > cleanup']         .add(configure__migration_cleanup).when('t__migrate')
      t['properties > init']           .add(properties_init)
      t['properties > post']           .add(properties_post)
      t['importpackages > init']       .add(configure__importpackages_init)
      t['database > init']             .add(configure__database_init)
      t['database > updatedatabase']   .add(t.t__database_updatedatabase)
      t['docparts > init']             .add(configure__docparts_init)
      t['docparts > migration']        .add(configure__docparts_migration).when('t__migrate')

      t.execute()
      //api.stop()
   }

   def configure__migration_init =
   {
      pass.with {
         _updateMetadata =
         t__migrate = getProperty('casemgmtmvac.migration.is_enabled', '0') { toBoolean(it) }
         if (t__migrate)
         {
            switch (api.'config.migration.sas.version')
            {
            case '9.2':
               _oldServerConfig = canonical("${api.'config.lev.applications.dir'}/SASCaseManagementServerCfg")
               break
            case '9.3':
               t__searchandcall =
               [
                  actions: [ __template.t__deletemetadata ],
                  fqids:
                  [
                     [
                        type:    'Tree',
                        search:  "@Name = 'Ent Case Mgmt Configuration ${api.'casemgmtmvac.maintenance.from.version'}'"
                     ]
                  ]
               ]

               _oldServerConfig = canonical("${api.'config.lev.applications.dir'}/SASCaseManagementServerCfg")
               break
            case '9.4':
               _oldServerConfig = null
               break
            default:
               _oldServerConfig = null
               break
            }
         }
      }
   }

   def configure__migration_cleanup =
   {
      if (pass._oldServerConfig)
         new File(pass._oldServerConfig).deleteDir()
   }

   def configure__importpackages_init =
   {
      pass.with {
         if (t__migrate)
         {
            switch (api.'config.migration.sas.version')
            {
               case '9.2':
                  break
               case '9.4':
                  break
               default:
                  t__importpackages_migration_createfolders =
                  [
                     _sasstpFolder
                  ]

                  t__importpackages_migration_run =
                  [
                     packages:
                     [
                        [
                        name:          'ECMStoredProcesses',
                        packagesDir:   canonical("$_installConfigDir/Deployment/Packages/"),
                        folder:        _sasstpFolder,
                        tokens:        _filters
                        ]
                     ]
                  ]
                  break
            }
         }
      }
   }

   def configure__database_init =
   {
      pass.with {
         Boolean initDB = !(getProperty('casemgmtmvac.dbms.bypass_init', '0') { toBoolean(it) })
         String migratedVersion = getProperty('casemgmtmvac.maintenance.from.version', _releaseNumber)

         if (initDB)
         {
            t__database_updatedatabase =
            [
               databaseType:  _dbmsEngine,
               url:           api.'casemgmtmvac.dbms.url',
               driverClass:   api.'casemgmtmvac.dbms.driver_class',
               userid:        api.'casemgmtmvac.dbms.userid',
               passwd:        api.'casemgmtmvac.dbms.passwd',
               classpath:
               [
                  [
                     delimiter:  '|',
                     path:       [api.'casemgmtmvac.dbms.jar']
                  ],
               ],
               update:
               [
                  [
                     sql:
                     [
                        "$_configInstallDir/$_dbmsDirLabel/load_ddl.sql",
                        "$_configInstallDir/load_install_data.sql"
                     ]
                  ]
               ]
            ]
         }
         else if (t__migrate && (migratedVersion != _releaseNumber))
         {
            updateconfigure__database_init()
         }
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def configure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
         actions:
            [
               [replace:[
                  token:'instructions.casemgmtmvac.head',
                  value:api.'instructions_casemgmtmvac_head.txt']],
               [replace:[
                  token:'instructions.casemgmtmvac.details.table.head',
                  value:api.'instructions_casemgmtmvac_details_table_head.txt']],
               [prepend:[
                  token:'instructions.casemgmtmvac.details.table.item',
                  value:api.'instructions_casemgmtmvac_doc.txt']],
               [prepend:[
                  token:'instructions.casemgmtmvac.details.table.item',
                  value:api.'instructions_casemgmtmvac_dbms.txt']],
               [prepend:[
                  token:'instructions.casemgmtmvac.details.table.item',
                  value:api.'instructions_casemgmtmvac_workdir.txt']],
               [replace:[
                  token:'instructions.casemgmtmvac.details.table.tail',
                  value:api.'instructions_casemgmtmvac_details_table_tail.txt']],
            ]
         ]
      }
   }

   def configure__docparts_migration =
   {
      pass.with {
         t__docparts.actions +=
         [replace:[
            token:'instructions.casemgmtmvac.migration.report',
            value: api.'migrateconfig_header_intro.txt']
         ]
      }
   }

   def sasjobs_createuser =
   {
      api.createUserAddToGroups(pass.t__sasjobs_run)
   }

   def sasjobs_removeuser =
   {
      api.removeUserFromGroupsAndDelete(pass.t__sasjobs_run)
   }

   // ===================================================== TARGET: updateConfigure

   /**
      Target: updateConfigure

      @see com.sas.gate.api.template.ConfigureTemplate
   */
   def updateConfigure()
   {
      //api.stop()

      String maintenanceTo   = getProperty('casemgmtmvac.maintenance.to.version', '')
      String maintenanceFrom = getProperty('casemgmtmvac.maintenance.from.version', '')

      api.createProperty('casemgmtmvac.maint.from.release', "${api.baseVersion(maintenanceFrom)}")

      if (maintenanceFrom)
      {
         if (  (maintenanceTo != maintenanceFrom) &&
               (api.baseVersion(maintenanceTo) != api.baseVersion(maintenanceFrom)))
         {
            UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

            t['properties > pre']            .add(updateconfigure__pre_init)
            t['properties > init']           .add(properties_init)
            t['properties > post']           .add(properties_post)
            t['loadmetadata > init']         .add(updateconfigure__loadmetadata_init)
            t['setobjectproperties > init']  .add(updateconfigure__setobjectproperties_init)
            t['database > init']             .add(updateconfigure__database_init)
            t['docparts > init']             .add(updateconfigure__docparts_init)

            t.execute()
         }
         else
         {
            UpdateConfigureTemplate t = new UpdateConfigureTemplate(this)

            t['properties > init']        .add(properties_init)
            t['properties > post']        .add(properties_post)
            t['docparts > init']          .add(updateconfigure__docparts_l10n_init)

            t.execute()
         }
      }
      else
         configure()

     //api.stop()
   }

   def updateconfigure__pre_init =
   {
      String legalName = api.createProperty('casemgmtmvac.short.legal.name', 'Ent Case Mgmt Server')

      String version = api.baseVersion(api.'casemgmtmvac.maintenance.to.version', api.'casemgmtmvac.release.number')
      api.createProperty('casemgmtmvac.release.number', version)

      String softwareComponent = api.createProperty('casemgmtmvac.software.component.name', "$legalName $version")
      String applicationFolder = api.'casemgmtmvac.application.folder'
      String oldFolderName = "$applicationFolder/$legalName 6.1"

      String configDir = "${new File(api.'casemgmtmvac.config.dir').parent}/$version"
      api.createProperty('casemgmtmvac.config.dir', configDir)
      String subFolder = api.createProperty('casemgmtmvac.application.sub.folder', "$applicationFolder/$softwareComponent")

      api.createProperty('casemgmtmvac.application.stp.folder', "$subFolder/Application SAS Code")

      pass.with
      {
         t__removemetadatabypath =
         [
            paths:
            [
               [
                  objectPath     : oldFolderName,
                  deleteContents : true
               ],
            ]
         ]
      }
   }

   def updateconfigure__loadmetadata_init =
   {
      pass.with {
         t__loadmetadata_update =
         [
            file: canonical("$_configMetadataDir/casemgmtmvac.appxml"),
            folder: _applicationSubFolder,
            applicationname: _softwareComponentName
         ]
      }
   }

   def updateconfigure__setobjectproperties_init =
   {
      pass.with {
         t__setobjectproperties +=
         [
            existingproperties: 'replace'
         ]
      }
   }

   def updateconfigure__database_init =
   {
      // TODO: Fill out with versions prior to 6.1
      String dbUpdate =
      [
         db2:
         [
            '3.2':   'dbpatch_3_2_db2.sql',
            '6.1':   'dbpatch_6_1_db2.sql',
         ],
         oracle:
         [
            '2.3':   'dbpatch_2_3_oracle.sql',
            '3.1':   'dbpatch_3_1_oracle.sql',
            '3.2':   'dbpatch_3_2_oracle.sql',
            '6.1':   'dbpatch_6_1_oracle.sql',
         ],
         postgres:
         [
            '2.3':   'dbpatch_2_3_postgresql.sql',
            '3.1':   'dbpatch_3_1_postgresql.sql',
            '3.2':   'dbpatch_3_2_postgresql.sql',
            '6.1':   'dbpatch_6_1_postgresql.sql',
         ],
         sqlsvr:
         [
            '2.3':   'dbpatch_2_3_sqlserver.sql',
            '3.1':   'dbpatch_3_1_sqlserver.sql',
            '3.2':   'dbpatch_3_2_sqlserver.sql',
            '6.1':   'dbpatch_6_1_sqlserver.sql',
         ],
         odbc:
         [
            '2.3':   'dbpatch_2_3_sqlserver.sql',
            '3.1':   'dbpatch_3_1_sqlserver.sql',
            '3.2':   'dbpatch_3_2_sqlserver.sql',
            '6.1':   'dbpatch_6_1_sqlserver.sql',
         ]
      ]?."${pass._dbmsEngine}"?."${pass._maintenanceFrom}"

      if (dbUpdate)
      {
         pass.with {
            t__database_updatedatabase =
            [
               databaseType   : _dbmsEngine,
               url            : api.'casemgmtmvac.dbms.url',
               driverClass    : api.'casemgmtmvac.dbms.driver_class',
               userid         : api.'casemgmtmvac.dbms.userid',
               passwd         : api.'casemgmtmvac.dbms.passwd',
               classpath      : [
                                   [
                                       delimiter   : '|',
                                       path        :  [api.'casemgmtmvac.dbms.jar']
                                   ],
                                ],
               update:
               [
                  [
                     sql: ["$_configInstallDir/$_dbmsDirLabel/$dbUpdate"],
                  ],
               ]
            ]
         }
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def updateconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [append:[
                  token:'instructions.casemgmtmvac.head',
                  value:api.'updateconfig_header_intro.txt']],
               [replace:[
                  token:'instructions.casemgmtmvac.head',
                  value:api.'instructions_casemgmtmvac_head.txt']],
               [prepend:[
                  token:'instructions.casemgmtmvac.details.table.head',
                  value:api.'instructions_casemgmtmvac_details_table_head.txt']],
               [prepend:[
                  token:'instructions.casemgmtmvac.details.table.item',
                  value:api.'instructions_casemgmtmvac_doc.txt']],
               [replace:[
                  token:'instructions.casemgmtmvac.details.table.tail',
                  value:api.'instructions_casemgmtmvac_details_table_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: checkForErrors

   /**
      Target: checkForErrors

      @see com.sas.gate.api.template.CheckForErrorsTemplate
   */
   def checkForErrors()
   {
      CheckForErrorsTemplate t = new CheckForErrorsTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: reportFailures

   /**
      Target: reportFailures

      @see com.sas.gate.api.template.ReportFailuresTemplate
   */
   def reportFailures()
   {
      ReportFailuresTemplate t = new ReportFailuresTemplate(this)
      t.execute()
   }

   // ===================================================== TARGET: unconfigure

   /**
      Target: unconfigure

      @see com.sas.gate.api.template.UnconfigureTemplate
   */
   def unconfigure()
   {
      //api.stop()

      UnconfigureTemplate t = new UnconfigureTemplate(this)

      t['properties > init']  .add(properties_init)
      t['properties > post']  .add(properties_post)
      t['init']               .add(unconfigure__init)
      t['docparts > init']    .add(unconfigure__docparts_init)

      t.execute()

      //api.stop()
   }

   def unconfigure__init =
   {
      pass.with
      {
         t__removemetadatabypath =
         [
            paths:
            [
               [
                  objectPath     : _applicationFolder,
                  deleteContents : true
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>properties &gt; post</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in ConfigurationRemoval.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>file</key>- <type>String</type>: File to replace tokens in.<br>
            <key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def unconfigure__docparts_init =
   {
      pass.with
      {
         t__docparts =
         [
            actions:
            [
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.casemgmtmvac.head',
                  value:   api.'unconfiguration_summary_casemgmtmvac_head.txt']],
               [prepend:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.casemgmtmvac.item',
                  value:   api.'unconfiguration_summary_casemgmtmvac_item.txt']],
               [replace:[
                  file:    api.'config.lev.unconfiguration.instructions.file',
                  token:   'unconfiguration.summary.casemgmtmvac.tail',
                  value:   api.'unconfiguration_summary_casemgmtmvac_tail.txt']],
            ]
         ]
      }
   }

   // ===================================================== TARGET: changeHost

   def changeHost()
   {
      //api.stop(properties:true)
      ChangeHostTemplate t = new ChangeHostTemplate(this)

      t['properties > init']  .add(properties_init)
      t['properties > post']  .add(properties_post)
      t['changehost > init']  .add(changehost__init)
      t['docparts > init']    .add(changehost__docparts_init)

      t.execute()
      //api.stop(properties:true)
   }

   def changehost__init =
   {
      pass.with {
         t__changehost =
         [
            exact:      ['casemgmtmvac.dbms.host'],
            substring:  ['casemgmtmvac.dbms.url'],
            updatefiles:
            [
               [
                  preToken: 'metaserver="',
                  file:     _autoexecPath,
               ],
               [
                  preToken: "ecm_dbhost='",
                  file:     _autoexecPath,
               ],
            ]
         ]
      }
   }

   /**
      Callback for hook: <hook>docparts &gt; init</hook>

      @param t__docparts   <type>ArrayList&lt;Map&gt;</type>: tokens in Instructions.html to replace and properties to replace them with
            <indent><key>replace, prepend, append</key>- <type>String</type>: Type of replacement to perform<br>
            <indent><key>token</key>- <type>String</type>: Token in Instructions.html to replace<br>
            <key>value</key>- <type>String</type>: Property to replace token with</indent></indent>

   */
   def changehost__docparts_init =
   {
      pass.with
      {
         if (t__changehost.docparts)
         {
            t__docparts =
            [
               file:    api.'config.lev.changehost.file',
               actions: []
            ]

            if (t__changehost.docparts.substring)
            {
               def (oldProperties, updatedProperties) = t__changehost.docparts.substring
               oldProperties.each { k, v ->

                  def value = new PipeCode([api.'changeHost_metadata_object_changed.txt'])
                     .tokenstringsub(tokens:
                     [
                        '_!PROPERTY_NAME':   k,
                        '_!OLD_VALUE':       v,
                        '_!NEW_VALUE':       updatedProperties."$k",
                     ], prefix:'${', suffix:'}').output[0]

                  t__docparts.actions += [prepend:[token: 'changeHost.configobj.item', value:value]]
               }
            }
         }
      }
   }
}
